import responses

from eveuniverse.models import EveType

from allianceauth.utils.testing import NoSocketsTestCase

from metenox.api.fuzzwork import get_type_ids_prices
from metenox.models import EveTypePrice
from metenox.tests.testdata.load_eveuniverse import load_eveuniverse


class TestFuzzWork(NoSocketsTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_crash_empty_search_query(self):
        """
        With an empty search query the API returns [] instead of a dict.
        This makes the code crash whe calling .items() on the result
        """

        get_type_ids_prices([])

    def test_no_price_update_on_zero(self):
        """
        If the new price of an eve type is zero it is refused
        """

        eve_type = EveType.objects.get(id=16634)
        type_price = EveTypePrice.objects.create(eve_type=eve_type, price=10_000)

        type_price.update_price(0)

        self.assertEqual(type_price.price, 10_000)

    @responses.activate
    def test_price_update(self):
        """
        The update_price method of EveTypePrice was crashing because fuzzwork api returned strings insteaf of float
        """

        eve_type = EveType.objects.get(id=16634)
        type_price = EveTypePrice.objects.create(eve_type=eve_type, price=100)

        responses.add(
            responses.GET,
            "https://market.fuzzwork.co.uk/aggregates/?region=10000002&types=16634",
            json={
                "16634": {
                    "buy": {
                        "weightedAverage": "3.6472570077758157",
                        "max": "6.0",
                        "min": "1.0",
                        "stddev": "1.0075961492582244",
                        "median": "4.01",
                        "volume": "9217403417.0",
                        "orderCount": "50",
                        "percentile": "4.48628743026839",
                    },
                    "sell": {
                        "weightedAverage": "7.128700104959353",
                        "max": "49990.0",
                        "min": "4.59",
                        "stddev": "5660.896280924974",
                        "median": "5.625",
                        "volume": "8841046323.0",
                        "orderCount": "78",
                        "percentile": "5.0194735444105",
                    },
                }
            },
            status=200,
        )

        new_price = get_type_ids_prices([16634])
        self.assertEqual(new_price[16634], 4.48628743026839)

        type_price.update_price(new_price[16634])

        self.assertEqual(type_price.price, 4.48628743026839)



"""Price views"""

from django.contrib.auth.decorators import login_required
from django.shortcuts import render

from metenox.models import EveTypePrice
from metenox.views.general import add_common_context


@login_required
def prices(request):
    """Displays moon goo prices"""
    goo_prices = EveTypePrice.objects.all()

    return render(
        request,
        "metenox/prices.html",
        add_common_context(
            {
                "goo_prices": goo_prices,
            }
        ),
    )



# Generated by Django 4.2.15 on 2024-10-09 20:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("metenox", "0003_remove_metenox_tag_metenox_tags_metenoxtag_default"),
    ]

    operations = [
        migrations.AlterField(
            model_name="metenox",
            name="tags",
            field=models.ManyToManyField(
                blank=True,
                help_text="Tags assigned to a Metenox",
                to="metenox.metenoxtag",
            ),
        ),
    ]



from django.core.management.base import BaseCommand

from allianceauth.services.hooks import get_extension_logger

from metenox import tasks

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = "Update all metenox holdings"

    def handle(self, *args, **options):
        logger.info("Initiating holding update tasks")
        tasks.update_all_holdings.delay()



from django.test import TestCase

from metenox.models import Webhook
from metenox.tests.testdata.load_eveuniverse import load_eveuniverse
from metenox.tests.utils import create_test_holding


class TestWebhooks(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_parse_webhook(self):
        """Parses a correct webhook and gets information"""

        holding_corporation = create_test_holding()

        webhook_url = "https://discord.com/api/webhooks/1105231532688192/1uR8aOH95dvZoGUpIdqwKed2z3R2rwH-JNdwDw2hP_IaYeuJnDDdwq41IBfQkBm5GKR_AApfv"

        webhook = Webhook.create_from_url(webhook_url, holding_corporation)

        self.assertIsNotNone(webhook)
        self.assertEqual(webhook.webhook_id, 1105231532688192)
        self.assertEqual(
            webhook.webhook_token,
            "1uR8aOH95dvZoGUpIdqwKed2z3R2rwH-JNdwDw2hP_IaYeuJnDDdwq41IBfQkBm5GKR_AApfv",
        )
        self.assertIn(holding_corporation, webhook.holding_corporations.all())
        self.assertEqual(webhook, Webhook.objects.all()[0])
        self.assertEqual(webhook.name, "")

    def test_parse_webhook_with_name(self):
        """Parses a correct webhook and its name then gets information"""

        holding_corporation = create_test_holding()

        webhook_url = "https://discord.com/api/webhooks/1105231532688192/1uR8aOH95dvZoGUpIdqwKed2z3R2rwH-JNdwDw2hP_IaYeuJnDDdwq41IBfQkBm5GKR_AApfv"

        webhook = Webhook.create_from_url(webhook_url, holding_corporation, "test_name")

        self.assertIsNotNone(webhook)
        self.assertEqual(webhook.webhook_id, 1105231532688192)
        self.assertEqual(
            webhook.webhook_token,
            "1uR8aOH95dvZoGUpIdqwKed2z3R2rwH-JNdwDw2hP_IaYeuJnDDdwq41IBfQkBm5GKR_AApfv",
        )
        self.assertIn(holding_corporation, webhook.holding_corporations.all())
        self.assertEqual(webhook, Webhook.objects.all()[0])
        self.assertEqual(webhook.name, "test_name")

    def test_parse_webhook_error(self):
        """Parses an incorrect webhook that will return a None"""

        holding_corporation = create_test_holding()

        webhook_url = "https://discord.co/adwpi/webhok/1105231532688192/1uR8aOH95dvZoGUpIdqwKed2z3R2rwH-JNdwDw2hP_IaYeuJnDDdwq41IBfQkBm5GKR_AApfv"

        webhook = Webhook.create_from_url(webhook_url, holding_corporation)

        self.assertIsNone(webhook)

    def test_build_url(self):
        """Checks that URLs are built as expected"""

        holding_corporation = create_test_holding()

        webhook_url = "https://discord.com/api/webhooks/1105231532688192/1uR8aOH95dvZoGUpIdqwKed2z3R2rwH-JNdwDw2hP_IaYeuJnDDdwq41IBfQkBm5GKR_AApfv"

        webhook = Webhook.create_from_url(webhook_url, holding_corporation)

        self.assertEqual(webhook.url, webhook_url)

    def test_add_default_webhooks(self):
        """Checks if default webhooks are properly applied"""

        webhook1 = Webhook.objects.create(
            webhook_id=1, webhook_token="a", name="webhook1", default=True
        )
        webhook2 = Webhook.objects.create(
            webhook_id=2, webhook_token="b", name="webhook2", default=True
        )
        webhook3 = Webhook.objects.create(
            webhook_id=3, webhook_token="c", name="webhook3", default=False
        )

        holding_corporation = create_test_holding()

        Webhook.add_default_webhooks_to_corporation(holding_corporation)

        corporation_webhooks = holding_corporation.webhooks.all()

        self.assertIn(webhook1, corporation_webhooks)
        self.assertIn(webhook2, corporation_webhooks)
        self.assertNotIn(webhook3, corporation_webhooks)



"""Models."""

import datetime
import re
from datetime import timedelta
from math import ceil, floor
from typing import Dict, List, Optional, Set

import dhooks_lite
from moonmining.models import Moon as MoonminigMoon

from django.contrib.auth.models import User
from django.contrib.humanize.templatetags import humanize
from django.db import models
from django.db.models import Count, Min, Sum
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EveGroup, EveMoon, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.evelinks.dotlan import solar_system_url
from allianceauth.eveonline.evelinks.eveimageserver import type_icon_url
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.notifications.models import Notification
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins

from metenox.app_settings import (
    METENOX_ADMIN_NOTIFICATIONS_ENABLED,
    METENOX_DAYS_BETWEEN_GOO_BAY_PINGS,
    METENOX_FUEL_BLOCKS_PER_HOUR,
    METENOX_MAGMATIC_GASES_PER_HOUR,
    METENOX_MOON_MATERIAL_BAY_CAPACITY,
)
from metenox.templatetags.metenox import formatisk

ESI_SCOPES = [
    "esi-universe.read_structures.v1",
    "esi-corporations.read_structures.v1",
    "esi-assets.read_corporation_assets.v1",
]

logger = get_extension_logger(__name__)


class General(models.Model):
    """A meta model for app permissions."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("view_moons", "Can see all scanned moons"),
            (
                "view_metenoxes",
                "Can add owners and view related corporations metenoxes",
            ),
            ("corporation_manager", "Can modify a corporation ping settings"),
            ("auditor", "Can access metenox information about all corporations"),
        )


class HoldingCorporation(models.Model):
    """Corporation holding metenox moon drills"""

    corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE, primary_key=True
    )

    is_active = models.BooleanField(default=True)
    last_updated = models.DateTimeField(null=True, default=None)

    ping_on_remaining_magmatic_days = models.IntegerField(
        default=0,
        help_text="Ping should be sent when the magmatic gases stored in a Metenox allow less than this value",
    )
    ping_on_remaining_fuel_days = models.IntegerField(
        default=0,
        help_text="Ping should be sent when the fuel blocks stored in a Metenox allow less than this value",
    )
    ping_on_value_in_moongoo_bay = models.IntegerField(
        null=True,
        blank=True,
        help_text=_(
            "Ping should be sent when the value stored in the moon bay of a Metenox exceed this value"
        ),
    )
    ping_on_volume_in_moongoo_bay = models.IntegerField(
        null=True,
        blank=True,
        help_text=_(
            "Ping should be sent when the volume stored in the moon goo bay of a Metenox exceed this value."
        ),
    )

    @property
    def alliance(self):
        """Returns holding corp's alliance"""
        return self.corporation.alliance

    @property
    def corporation_name(self):
        """Returns the holding corporation's name"""
        return self.corporation.corporation_name

    @property
    def count_metenoxes(self) -> int:
        """Return the number of metenoxes a holding corporation owns"""
        return self.metenoxes.count()

    @property
    def raw_revenue(self) -> float:
        """Returns the raw metenox revenues before fuel prices"""
        return self.metenoxes.aggregate(Sum("moon__value"))["moon__value__sum"]

    @property
    def profit(self) -> float:
        """Returns the metenoxes profit after fuel prices"""
        if self.raw_revenue:
            return self.raw_revenue - self.count_metenoxes * Moon.fuel_price()
        return 0.0

    def active_owners(self):
        """Returns corporation owners that haven't been disabled"""
        return self.owners.filter(is_enabled=True)

    def set_update_time_now(self):
        """Informs that this corporation has just been properly updated"""
        self.last_updated = timezone.now()
        self.save()

    def ping_metenox_fuel(self, metenox: "Metenox", new_fuel_blocks_amount: int):
        """Sends out to all webhooks a notification about the Metenox low fuel blocks level"""
        self.alert_webhooks(
            "Low fuel blocks level",
            f"Your metenox {metenox.structure_name} in "
            f"[{metenox.system_name}]({solar_system_url(metenox.system_name)}) "
            f"level of fuel blocks is under the threshold. Refueling is required",
            [("Remaining Fuel Blocks", new_fuel_blocks_amount)],
        )

    def ping_metenox_magma(self, metenox: "Metenox", new_magmatic_gases_amount: int):
        """Sends out to all webhooks a notification about the Metenox low magmatic level"""
        self.alert_webhooks(
            "Low reagent level",
            f"Your metenox {metenox.structure_name} in "
            f"[{metenox.system_name}]({solar_system_url(metenox.system_name)}) "
            f"level of magmatic gases is under the threshold. Refueling is required",
            [("Remaining Magmatic Gases", new_magmatic_gases_amount)],
        )

    def ping_metenox_value(
        self, metenox: "Metenox", current_moon_material_bay_value: float
    ):
        """Sends out to all webhooks a notification about the Metenox moon material bay value exceeding the threshold"""
        self.alert_webhooks(
            "Moon material bay value above threshold",
            f"Your metenox {metenox.structure_name} in "
            f"[{metenox.system_name}]({solar_system_url(metenox.system_name)}) "
            f"moon material bay value is over the threshold.",
            [
                ("Current value", f"{formatisk(current_moon_material_bay_value)} ISK"),
                (
                    "Threshold value",
                    f"{formatisk(self.ping_on_value_in_moongoo_bay)} ISK",
                ),
            ],
        )

    def ping_metenox_volume(
        self, metenox: "Metenox", current_moon_material_bay_volume: float
    ):
        """
        Sends out to all webhooks a notification about the Metenox moon material bay volume exceeding the threshold
        """
        self.alert_webhooks(
            "Moon material bay volume above threshold",
            f"Your metenox {metenox.structure_name} in "
            f"[{metenox.system_name}]({solar_system_url(metenox.system_name)}) "
            f"moon material bay volume is over the threshold.",
            [
                (
                    "Current volume",
                    f"{humanize.intcomma(current_moon_material_bay_volume)} m3",
                ),
                (
                    "Threshold value",
                    f"{humanize.intcomma(self.ping_on_volume_in_moongoo_bay)} m3",
                ),
            ],
        )

    def alert_webhooks(self, title, message, fields):
        """Sends an alert to all metenoxes linked to a corporation"""
        logger.info(
            "Sending a ping to every webhook of corporation id %s",
            self.corporation.corporation_id,
        )
        for webhook in self.webhooks.all():
            webhook.send_alert(
                title,
                message,
                fields,
                self.corporation_name,
                self.corporation.logo_url(32),
            )

    def get_holding_tags(self) -> List["MetenoxTag"]:
        """Return all tags used for metenoxes of this holding corporation"""
        metenoxes = self.metenoxes.all()
        return list(MetenoxTag.objects.filter(metenoxes__in=metenoxes))

    def count_metenoxes_with_tag(self, tag: "MetenoxTag") -> int:
        """Count of metenoxes in the holding corporation with this tag"""
        return Metenox.objects.filter(tags=tag, corporation=self).count()

    def profit_metenoxes_with_tag(self, tag: "MetenoxTag") -> float:
        """Return the profit of all metenoxes in the holding corporation with this tag"""
        return Metenox.objects.filter(tags=tag, corporation=self).aggregate(
            Sum("moon__value")
        )["moon__value__sum"]

    def __str__(self) -> str:
        return self.corporation_name

    @classmethod
    def enable_all_owners(cls, queryset):
        """Enable all the owners of the holding corporations listed in the queryset"""
        owners_qs = Owner.objects.filter(corporation__in=queryset)
        Owner.enable_owners(owners_qs)

    @classmethod
    def disable_all_owners(cls, queryset):
        """Disable all the owners of the holding corporations listed in the queryset"""
        owners_qs = Owner.objects.filter(corporation__in=queryset)
        Owner.disable_owners(owners_qs)


class Owner(models.Model):
    """Character in corporation owning metenoxes"""

    corporation = models.ForeignKey(
        HoldingCorporation,
        on_delete=models.CASCADE,
        related_name="owners",
    )

    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        related_name="+",
        help_text="Character used to sync this corporation from ESI",
    )

    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text="Disabled corporations are excluded from the update process",
    )

    class Meta:
        verbose_name = "Owner"
        verbose_name_plural = "Owners"

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        """Return name."""
        alliance_ticker_str = (
            f" [{self.corporation.alliance.alliance_ticker}]"
            if self.corporation.alliance
            else ""
        )
        return f"{self.corporation}{alliance_ticker_str} - {self.character_name}"

    @property
    def character_name(self) -> str:
        """
        Returns the character name of this owner.
        If it detects that the ownership was deleted it will delete the owner.
        """
        if not self.character_ownership:
            self.delete()  # Will automatically cleanup the dangling toon for the next time
            return "No char associated - owner is being cleaned up"
        return str(self.character_ownership.character)

    @property
    def alliance_name(self) -> str:
        """Return alliance name."""
        return (
            self.corporation.alliance.alliance_name if self.corporation.alliance else ""
        )

    @property
    def user(self) -> User:
        """Return the user linked to this owner"""
        return self.character_ownership.user

    def fetch_token(self) -> Token:
        """Return valid token for this mining corp or raise exception on any error."""
        if not self.character_ownership:
            raise RuntimeError("This owner has no character configured.")
        token = (
            Token.objects.filter(
                character_id=self.character_ownership.character.character_id
            )
            .require_scopes(ESI_SCOPES)
            .require_valid()
            .first()
        )
        if not token:
            raise Token.DoesNotExist(f"{self}: No valid token found.")
        return token

    def enable(self):
        """Enable an owner after logging in the toon again"""
        self.is_enabled = True
        self.save()

    def disable(self, notify=True, cause="Unspecified"):
        """Disables the owner after an ESI error"""
        if notify:
            level = Notification.Level.WARNING
            title = "Metenox owner disabled"
            message = f"The owner {self} has been disabled because of {cause}"
            Notification.objects.notify_user(
                self.user,
                title,
                message,
                level,
            )
            if METENOX_ADMIN_NOTIFICATIONS_ENABLED:
                notify_admins(message, title, level)
        self.is_enabled = False
        self.save()

    @classmethod
    def get_owners_associated_to_user(cls, user: User):
        """Returns all owners of the user"""
        return cls.objects.filter(character_ownership__user=user)

    @classmethod
    def enable_owners(cls, queryset):
        """Receives a queryset of Owners and sets all of them as enabled"""
        queryset.update(is_enabled=True)

    @classmethod
    def disable_owners(cls, queryset):
        """Receives a queryset of Owners and sets all of them as disabled"""
        queryset.update(is_enabled=False)


class Moon(models.Model):
    """Represents a moon and the metenox related values"""

    eve_moon = models.OneToOneField(
        EveMoon, on_delete=models.CASCADE, primary_key=True, related_name="+"
    )

    moonmining_moon = models.OneToOneField(
        MoonminigMoon,
        on_delete=models.CASCADE,
        null=True,  # metenox might be dropped on an unknown moon
        default=None,
        related_name="+",
    )

    value = models.FloatField(default=0)
    value_updated_at = models.DateTimeField(null=True, default=None)

    @property
    def hourly_pull(self) -> Dict[EveType, int]:
        """Returns how much goo is harvested in an hour by a metenox"""
        hourly_products = MetenoxHourlyProducts.objects.filter(moon=self)
        return {product.product: product.amount for product in hourly_products}

    @property
    def name(self) -> str:
        """Returns name of this moon"""
        return self.eve_moon.name.replace("Moon ", "")

    @property
    def rarity_class(self):
        """Returns rarity class of this moon"""
        return self.moonmining_moon.rarity_class

    def update_price(self):
        """Updates the Metenox price attribute to display"""
        hourly_harvest_value = sum(
            EveTypePrice.get_eve_type_price(moon_goo) * moon_goo_amount
            for moon_goo, moon_goo_amount in self.hourly_pull.items()
        )
        self.value = hourly_harvest_value * 24 * 30
        self.value_updated_at = timezone.now()
        self.save()

    @property
    def cycles_before_full(self) -> int:
        """Number of harvest cycles before the moon material bay is at full capacity"""
        bay_capacity = METENOX_MOON_MATERIAL_BAY_CAPACITY
        harvest_volume = sum(
            goo_type.volume * amount for goo_type, amount in self.hourly_pull.items()
        )
        return ceil(bay_capacity / harvest_volume) if harvest_volume != 0 else 0

    @classmethod
    def fuel_price(cls) -> float:
        """Returns the monthly price of running a metenox"""
        hourly_price = (
            METENOX_MAGMATIC_GASES_PER_HOUR * EveTypePrice.get_magmatic_gases_price()
            + METENOX_FUEL_BLOCKS_PER_HOUR * EveTypePrice.get_fuel_block_price()
        )
        return hourly_price * 24 * 30

    @classmethod
    def moons_in_need_of_update(cls):
        """
        Returns a queryset of all moons that have products in the moonmining application but none here
        """

        moons = Moon.objects.annotate(
            num_metenox_products=Count("hourly_products", distinct=True),
            num_moonmining_products=Count("moonmining_moon__products", distinct=True),
        )
        return moons.filter(num_metenox_products__lt=1, num_moonmining_products__gt=0)

    @property
    def profit(self) -> float:
        """Returns the monthly profit of a meteneox including the fuel price"""
        return self.value - self.fuel_price()

    def __str__(self):
        return self.name


class Metenox(models.Model):
    """
    Represents a metenox anchored on a moon
    """

    structure_id = models.PositiveBigIntegerField(primary_key=True)
    structure_name = models.TextField(max_length=150)
    tags = models.ManyToManyField(
        "MetenoxTag",
        blank=True,
        related_name="metenoxes",
        help_text="Tags assigned to a Metenox",
    )

    moon = models.OneToOneField(
        Moon,
        on_delete=models.CASCADE,
        related_name="metenox",
    )
    corporation = models.ForeignKey(
        HoldingCorporation, on_delete=models.CASCADE, related_name="metenoxes"
    )

    fuel_blocks_count = models.IntegerField(default=0)
    magmatic_gas_count = models.IntegerField(default=0)

    was_magmatic_pinged = models.BooleanField(
        default=False,
        help_text="If a ping has been sent out after noticing a low magmatic level",
    )
    was_fuel_pinged = models.BooleanField(
        default=False,
        help_text="If a ping has been sent out after noticing a low fuel level",
    )
    last_moongoo_bay_ping = models.DateTimeField(
        null=True,
        help_text=_(
            "Last time that a ping was sent about the moon goo bay exceeding the threshold"
        ),
    )

    last_updated = models.DateTimeField(
        auto_now=True,
        blank=True,
    )

    @property
    def system_name(self) -> str:
        """Returns the system name from the structure name"""
        return self.structure_name.split(" - ", maxsplit=1)[0]

    def get_stored_moon_materials(self) -> List["MetenoxStoredMoonMaterials"]:
        """Return the list of stored moon materials in that Metenox"""
        return list(self.stored_moon_materials.all())

    def get_stored_moon_materials_value(self) -> float:
        """Return the value of all moon materials stored in the metenox"""
        return sum(
            EveTypePrice.get_eve_type_price(stored_moon_material.product)
            * stored_moon_material.amount
            for stored_moon_material in self.get_stored_moon_materials()
        )

    def get_stored_moon_materials_volume(self) -> float:
        """Return the volume of all moon materials stored in the metenox"""
        return round(
            sum(
                stored_moon_material.product.volume * stored_moon_material.amount
                for stored_moon_material in self.get_stored_moon_materials()
            ),
            2,
        )

    def set_fuel_blocs(self, new_amount):
        """Edits the amount of fuel blocks in the Metenox and sends a notification if needed"""

        self.fuel_blocks_count = new_amount

        remaining_fuel_days = floor(
            self.fuel_blocks_count / (METENOX_FUEL_BLOCKS_PER_HOUR * 24)
        )

        ping_on_remaining_fuel_days = self.corporation.ping_on_remaining_fuel_days
        if (
            ping_on_remaining_fuel_days > remaining_fuel_days
            and not self.was_fuel_pinged
        ):
            logger.info(
                "Fuel blocks level of metenox id %s are under %s (%s)",
                self.structure_id,
                ping_on_remaining_fuel_days,
                new_amount,
            )
            self.corporation.ping_metenox_fuel(self, new_amount)
            self.was_fuel_pinged = True

        elif (
            ping_on_remaining_fuel_days <= remaining_fuel_days and self.was_fuel_pinged
        ):
            logger.info(
                "Fuel block level of metenox id %s is now back above %s (%s)",
                self.structure_id,
                ping_on_remaining_fuel_days,
                new_amount,
            )
            self.was_fuel_pinged = False

        self.save()

    def set_magmatic_gases(self, new_amount):
        """Edits the amount of magmatic gases in the Metenox and sends a notification if needed"""

        self.magmatic_gas_count = new_amount

        remaining_gas_days = floor(
            self.magmatic_gas_count / (METENOX_MAGMATIC_GASES_PER_HOUR * 24)
        )

        ping_on_remaining_magmatic_days = (
            self.corporation.ping_on_remaining_magmatic_days
        )

        if (
            ping_on_remaining_magmatic_days > remaining_gas_days
            and not self.was_magmatic_pinged
        ):
            logger.info(
                "Magmatic gas level of metenox id %s are under %s (%s)",
                self.structure_id,
                ping_on_remaining_magmatic_days,
                new_amount,
            )
            self.corporation.ping_metenox_magma(self, new_amount)
            self.was_magmatic_pinged = True

        elif (
            ping_on_remaining_magmatic_days <= remaining_gas_days
            and self.was_magmatic_pinged
        ):
            logger.info(
                "Magmatic gas level of metenox id %s is now back above %s (%s)",
                self.structure_id,
                ping_on_remaining_magmatic_days,
                new_amount,
            )
            self.was_magmatic_pinged = False

        self.save()

    def check_moon_material_bay(self):
        """Checks if a ping for the moon material bay should be sent out based on the current volume/value of the bay"""
        logger.info(
            "Checking if a discord ping should be sent for metenox id %d moon material bay",
            self.structure_id,
        )
        if self.last_moongoo_bay_ping:
            if not timezone.now() - self.last_moongoo_bay_ping > timedelta(
                days=METENOX_DAYS_BETWEEN_GOO_BAY_PINGS
            ):
                logger.info("Last ping was too recent. Aborting")
                return  # we already did a ping too recently

        ping_sent = False

        moon_material_bay_value = self.get_stored_moon_materials_value()
        ping_on_value = self.corporation.ping_on_value_in_moongoo_bay
        logger.debug(
            "Current value is at %s and limit is at %s",
            moon_material_bay_value,
            ping_on_value,
        )
        if ping_on_value and moon_material_bay_value > ping_on_value:
            logger.info("Sending out a ping for a value of %s", moon_material_bay_value)
            self.corporation.ping_metenox_value(self, moon_material_bay_value)
            ping_sent = True

        moon_material_bay_volume = self.get_stored_moon_materials_volume()
        ping_on_volume = self.corporation.ping_on_volume_in_moongoo_bay
        logger.debug(
            "Current volume is at %s and limit is at %s",
            moon_material_bay_volume,
            ping_on_volume,
        )
        if ping_on_volume and moon_material_bay_volume > ping_on_volume:
            logger.info(
                "Sending out a ping for a volume of %s", moon_material_bay_volume
            )
            self.corporation.ping_metenox_volume(self, moon_material_bay_volume)
            ping_sent = True

        if ping_sent:
            self.last_moongoo_bay_ping = timezone.now()
            self.save()

    def __str__(self):
        return self.structure_name

    class Meta:
        verbose_name_plural = "Metenoxes"


class MetenoxHourlyProducts(models.Model):
    """
    Represents how much moon goo a Metenox harvests in an hour
    """

    moon = models.ForeignKey(
        Moon, on_delete=models.CASCADE, related_name="hourly_products"
    )
    product = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    amount = models.IntegerField()

    def __str__(self):
        return f"{self.product.name} - {self.amount}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["moon", "product"], name="functional_pk_metenoxhourlyproduct"
            )
        ]

    @classmethod
    def all_moon_goo_ids(cls) -> Set[int]:
        """Returns all known moon goo ids in the database"""
        return set(cls.objects.values_list("product", flat=True).order_by("product_id"))


class MetenoxStoredMoonMaterials(models.Model):
    """
    Represents how much moon materials are currently stored inside a Metenox
    """

    metenox = models.ForeignKey(
        Metenox, on_delete=models.CASCADE, related_name="stored_moon_materials"
    )
    product = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")

    amount = models.IntegerField(default=0)

    def __str__(self):
        return f"{self.product.name} - {self.amount} in {self.metenox.structure_name}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["metenox", "product"], name="functional_pk_metenoxstoredproduct"
            )
        ]


class MetenoxTag(models.Model):
    """
    Tags assigned to a metenox for easy sorting
    """

    name = models.CharField(max_length=20, unique=True)
    default = models.BooleanField(
        default=False, help_text="This tag will be applied on every new metenox if true"
    )

    def __str__(self):
        return self.name


class EveTypePrice(models.Model):
    """
    Represent an eve type and its last fetched price
    """

    __MOON_GOOS_GROUP_ID = 427
    __FUEL_BLOCK_GROUP_ID = 1136
    __MAGMATIC_TYPE_ID = 81143

    eve_type = models.OneToOneField(EveType, on_delete=models.CASCADE, related_name="+")
    price = models.FloatField(default=0)
    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f"{self.eve_type.name} - {self.price} ISK"

    def update_price(self, new_price: float):
        """Updates the price of an item"""
        if new_price <= 0:
            return
        self.price = new_price
        self.save()

    @classmethod
    def get_eve_type_price(cls, eve_type: EveType) -> float:
        """Returns the price of an item"""
        type_price, _ = cls.objects.get_or_create(eve_type=eve_type)
        return type_price.price

    @classmethod
    def get_eve_type_id_price(cls, eve_type_id: int) -> float:
        """Returns the price of an item id"""
        return cls.get_eve_type_price(EveType.objects.get(id=eve_type_id))

    @classmethod
    def _get_type_ids_from_group(cls, group_id: int) -> Set[int]:
        """Fetches type ids from their group and returns it as a set"""
        group = EveGroup.objects.get(id=group_id)
        return set(group.eve_types.filter(published=True).values_list("id", flat=True))

    @classmethod
    def get_fuels_type_ids(cls) -> Set[int]:
        """Fetches the id of all 4 fuel blocks from their group and magmatic gas"""
        return cls._get_type_ids_from_group(cls.__FUEL_BLOCK_GROUP_ID) | {
            cls.__MAGMATIC_TYPE_ID
        }

    @classmethod
    def get_fuel_blocs_type_ids(cls) -> Set[int]:
        """Fetches the id of all 4 fuel blocks"""
        return cls._get_type_ids_from_group(cls.__FUEL_BLOCK_GROUP_ID)

    @classmethod
    def get_moon_goos_type_ids(cls) -> Set[int]:
        """Fetches the ids of all moon goos from their group"""
        return cls._get_type_ids_from_group(cls.__MOON_GOOS_GROUP_ID)

    @classmethod
    def get_magmatic_gas_type_id(cls) -> int:
        """Return the type of magmatic gases"""
        return cls.__MAGMATIC_TYPE_ID

    @classmethod
    def get_fuel_block_price(cls) -> float:
        """Returns the price of the cheapest fuel block"""
        return (
            cls.objects.filter(eve_type__eve_group=cls.__FUEL_BLOCK_GROUP_ID).aggregate(
                Min("price")
            )["price__min"]
            or 0.0
        )

    @classmethod
    def get_magmatic_gases_price(cls) -> float:
        """Returns the price of a unit of magmatic gases"""
        return cls.get_eve_type_id_price(cls.__MAGMATIC_TYPE_ID)


class Webhook(models.Model):
    """Represents a discord webhook information"""

    webhook_id = models.BigIntegerField(primary_key=True)
    webhook_token = models.TextField()

    name = models.CharField(
        max_length=30, unique=True, help_text="Text to recognize the webhook"
    )
    default = models.BooleanField(
        default=False,
        help_text="If the webhook will automatically be added to every new corporation",
    )

    holding_corporations = models.ManyToManyField(
        HoldingCorporation,
        related_name="webhooks",
        help_text="Corporation associated to this webhook",
    )

    @classmethod
    def create_from_url(
        cls,
        webhook_url: str,
        holding_corporation: HoldingCorporation,
        name: str = "",
    ) -> Optional["Webhook"]:
        """Parses a webhook URL to extract the information and returns the newly created webhook"""
        pattern = r"https:\/\/discord.com\/api\/webhooks\/(?P<id>\d+)\/(?P<token>.+)"
        match = re.match(pattern, webhook_url)

        if not match:  # regex didn't find the groups
            return None

        new_webhook = cls.objects.create(
            webhook_id=int(match["id"]),
            webhook_token=match["token"],
            name=name,
        )

        new_webhook.holding_corporations.add(holding_corporation)

        return new_webhook

    @classmethod
    def get_by_id(cls, webhook_id: int) -> Optional["Webhook"]:
        """Tries to find a webhook by its id"""
        try:
            return cls.objects.get(webhook_id=webhook_id)
        except cls.DoesNotExist:
            return None

    @property
    def url(self) -> str:
        """Builds the discord webhook URL"""
        return (
            f"https://discord.com/api/webhooks/{self.webhook_id}/{self.webhook_token}"
        )

    def _build_webhook(self) -> dhooks_lite.Webhook:
        """Builds the discord webhook and returns it"""
        return dhooks_lite.Webhook(
            self.url,
            username="Metenox notification",
            avatar_url="https://gitlab.com/uploads/-/system/project/avatar/60747747/oil-well-solid.png",
        )

    # pylint: disable = too-many-positional-arguments, too-many-arguments
    def send_alert(
        self,
        title: str,
        content: str,
        fields=None,
        author_name: str = None,
        author_icon_url: str = None,
    ):
        """Sends an alert through the webhook"""
        embed = self._build_embed(
            title, content, 0xFF5733, fields, author_name, author_icon_url
        )
        self._send_through_webhook([embed])

    # pylint: disable = too-many-positional-arguments, too-many-arguments
    def send_info(
        self,
        title: str,
        content: str,
        fields=None,
        author_name: str = None,
        author_icon_url: str = None,
    ):
        """Sends an information through the webhook"""
        embed = self._build_embed(
            title, content, 0x5CDBF0, fields, author_name, author_icon_url
        )
        self._send_through_webhook([embed])

    def _send_through_webhook(self, embeds: List[dhooks_lite.Embed]):
        """Sends embeds through the webhook"""
        webhook = self._build_webhook()
        webhook.execute(embeds=embeds)

    # pylint: disable = too-many-positional-arguments, too-many-arguments
    def _build_embed(
        self,
        title: str,
        content: str,
        color: int,
        fields=None,
        author_name: str = None,
        author_icon_url: str = None,
    ) -> dhooks_lite.Embed:
        """Build a discord embedd and returns it"""

        formatted_fields = []
        if fields:
            for field in fields:
                formatted_fields.append(dhooks_lite.Field(field[0], str(field[1])))
        if not author_name:
            author_name = "AA-Metenox"

        embed = dhooks_lite.Embed(
            author=dhooks_lite.Author(
                name=author_name,
                icon_url=author_icon_url,
            ),
            title=title,
            description=content,
            thumbnail=dhooks_lite.Thumbnail(type_icon_url(81826, size=128)),
            color=color,
            fields=formatted_fields,
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            footer=dhooks_lite.Footer(
                "aa-metenox",
            ),
        )

        return embed

    def add_corporation(self, corporation: HoldingCorporation):
        """Adds a corporation to the list that should be pinged"""
        self.holding_corporations.add(corporation)

    def remove_corporation(self, corporation: HoldingCorporation):
        """Removes a corporation from the list that should be pinged"""
        self.holding_corporations.remove(corporation)

    @classmethod
    def add_default_webhooks_to_corporation(cls, corporation: HoldingCorporation):
        """Adds all webhooks tagged as default to the corporation"""
        for webhook in cls.objects.filter(default=True):
            webhook.add_corporation(corporation)

    def __str__(self):
        return self.name



"""Admin site."""

from django.contrib import admin

from metenox.models import (
    EveTypePrice,
    HoldingCorporation,
    Metenox,
    MetenoxHourlyProducts,
    MetenoxStoredMoonMaterials,
    MetenoxTag,
    Moon,
    Owner,
    Webhook,
)
from metenox.templatetags.metenox import formatisk


class MetenoxHourlyHarvestInline(admin.TabularInline):
    model = MetenoxHourlyProducts

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Moon)
class MoonAdmin(admin.ModelAdmin):
    list_display = ["name", "moon_value", "value_updated_at"]
    fields = [("eve_moon", "moonmining_moon"), ("value", "value_updated_at")]
    search_fields = ["eve_moon__name"]
    readonly_fields = ["eve_moon", "moonmining_moon"]
    inlines = (MetenoxHourlyHarvestInline,)

    @admin.display(description="value")
    def moon_value(self, moon: Moon):
        return f"{formatisk(moon.value)} ISK"

    def has_add_permission(self, request):
        return False


@admin.action(description="Reset the metenox ping status")
def reset_pings(modeladmin, request, queryset):
    """Resets the ping status of the selected metenoxes for them to be pinged again"""
    queryset.update(was_magmatic_pinged=False, was_fuel_pinged=False)


class MetenoxMoonMaterialBayInline(admin.TabularInline):
    model = MetenoxStoredMoonMaterials

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Metenox)
class MetenoxAdmin(admin.ModelAdmin):
    list_display = [
        "structure_name",
        "corporation",
        "alliance",
        "tag_display",
        "metenox_value",
        "fuel_blocks_count",
        "magmatic_gas_count",
        "was_fuel_pinged",
        "was_magmatic_pinged",
        "last_moongoo_bay_ping",
        "last_updated",
    ]
    list_filter = ["corporation", "corporation__corporation__alliance", "tags"]
    search_fields = ["structure_name"]
    actions = [reset_pings]
    fields = [
        ("structure_id", "structure_name"),
        "moon",
        "corporation",
        "tags",
        ("fuel_blocks_count", "magmatic_gas_count"),
        ("was_fuel_pinged", "was_magmatic_pinged", "last_moongoo_bay_ping"),
        "last_updated",
    ]
    readonly_fields = [
        "structure_id",
        "moon",
        "corporation",
        "was_fuel_pinged",
        "was_magmatic_pinged",
        "last_moongoo_bay_ping",
        "last_updated",
    ]
    inlines = (MetenoxMoonMaterialBayInline,)

    @admin.display(description="Value", ordering="moon__value")
    def metenox_value(self, metenox: Metenox):
        return f"{formatisk(metenox.moon.value)} ISK"

    @admin.display(
        description="Corporation alliance",
        ordering="corporation__corporation__alliance",
    )
    def alliance(self, metenox: Metenox):
        return metenox.corporation.alliance

    @admin.display(description="Tags attached to the metenox")
    def tag_display(self, metenox: Metenox):
        return " ,".join([tag.name for tag in metenox.tags.all()])

    def has_add_permission(self, request):
        return False


@admin.action(description="Enable selected owners")
def enable_owners(modeladmin, request, queryset):
    """Enable all owners of the received queryset"""
    Owner.enable_owners(queryset)


@admin.action(description="Disable selected owners")
def disable_owners(modeladmin, request, queryset):
    """Disable all owners of the received queryset"""
    Owner.disable_owners(queryset)


@admin.register(Owner)
class OwnerAdmin(admin.ModelAdmin):
    list_display = [
        "char_name",
        "is_enabled",
        "corporation",
        "ally_name",
        "character_ownership",
    ]
    list_filter = ["is_enabled", "corporation", "corporation__corporation__alliance"]
    readonly_fields = ["character_ownership", "corporation"]
    actions = [enable_owners, disable_owners]

    @admin.display(
        description="Character name", ordering="character_ownership__character"
    )
    def char_name(self, owner: Owner) -> str:
        return owner.character_name

    @admin.display(
        description="Alliance name", ordering="corporation__corporation__alliance"
    )
    def ally_name(self, owner: Owner) -> str:
        return owner.alliance_name

    def has_add_permission(self, request):
        return False


@admin.action(description="Enable all owners of the corporations")
def enable_all_owners(modeladmin, request, queryset):
    """Enable all owners part of the corporations listed"""
    HoldingCorporation.enable_all_owners(queryset)


@admin.action(description="Disable all owners of the corporations")
def disable_all_owners(modeladmin, request, queryset):
    """Disable all owners part of the corporations listed"""
    HoldingCorporation.disable_all_owners(queryset)


class OwnerCharacterAdminInline(admin.TabularInline):
    model = Owner
    readonly_fields = ["character_ownership"]

    fields = ["character_ownership", "is_enabled"]

    def has_add_permission(self, request, obj):
        return False


@admin.register(HoldingCorporation)
class HoldingCorporationAdmin(admin.ModelAdmin):
    list_display = [
        "corporation",
        "alliance",
        "is_active",
        "count_metenox",
        "count_owners",
        "last_updated",
    ]
    list_filter = ["corporation__alliance", "is_active"]
    sortable_by = ["is_active"]
    actions = [enable_all_owners, disable_all_owners]
    readonly_fields = ["corporation", "last_updated", "count_metenox"]
    inlines = (OwnerCharacterAdminInline,)

    @admin.display(description="Number metenoxes")
    def count_metenox(self, holding: HoldingCorporation) -> int:
        return holding.count_metenoxes

    @admin.display(
        description="Number of owners / active owners", ordering="owners__count"
    )
    def count_owners(self, holding: HoldingCorporation) -> str:
        return f"{holding.owners.count()} / {holding.active_owners().count()}"

    def has_add_permission(self, request):
        return False


@admin.register(EveTypePrice)
class EveTypePriceAdmin(admin.ModelAdmin):
    list_display = ["eve_type", "eve_group", "type_price"]
    readonly_fields = ["eve_type", "last_update"]

    @admin.display(description="Price", ordering="price")
    def type_price(self, type_price: EveTypePrice):
        return f"{formatisk(type_price.price)} ISK"

    @admin.display(description="Eve Group", ordering="eve_type__eve_group")
    def eve_group(self, type_price: EveTypePrice):
        return type_price.eve_type.eve_group

    def has_add_permission(self, request):
        return False


@admin.action(description="Add the tags to every metenoxes")
def add_tags(modeladmin, request, queryset):
    for metenox in Metenox.objects.all():
        metenox.tags.add(*queryset)


@admin.register(MetenoxTag)
class MetenoxTagAdmin(admin.ModelAdmin):
    list_display = ["name", "default"]
    actions = [add_tags]


@admin.action(description="Send a test message to the selected webhooks")
def test_webhooks(modeladmin, request, queryset):
    """Test the selected webhooks"""
    for webhook in queryset:
        webhook.send_info(
            "Test notification",
            "This is a webhook test send from the aa-metenox application",
        )


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = ["name", "webhook_id", "count_holdings", "default"]
    list_filter = ["holding_corporations", "default"]
    actions = [test_webhooks]
    readonly_fields = ["webhook_id", "webhook_token"]

    @admin.display(
        description="Number of corporations attached to this webhook",
        ordering="webhooks__count",
    )
    def count_holdings(self, webhook: Webhook):
        return webhook.holding_corporations.count()



from django.core.management.base import BaseCommand

from allianceauth.services.hooks import get_extension_logger

from metenox import tasks

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = (
        "Fetches all moons from moonining application and adds them to this application"
    )

    def handle(self, *args, **options):
        tasks.update_moons_from_moonmining.delay()



"""App settings."""

from app_utils.app_settings import clean_setting

METENOX_ADMIN_NOTIFICATIONS_ENABLED = clean_setting(
    "METENOX_ADMIN_NOTIFICATION_ENABLED", True
)
"""Whether admins will get notifications about important events like
when someone adds a new owner.
"""

METENOX_MOON_MATERIAL_BAY_CAPACITY = clean_setting(
    "METENOX_MOON_MATERIAL_BAY_CAPACITY", 500_000
)
"""
Volume of the Metenox's Moon material Output Bay
Used to calculate how long a metenox takes before being full
"""

METENOX_HOURLY_HARVEST_VOLUME = clean_setting(
    "METENOX_HOURLY_HARVEST_VOLUME ",
    30_000,
)
"""
Hourly volume in m3 that a metenox will harvest.
This value shouldn't be edited
"""

METENOX_HARVEST_REPROCESS_YIELD = clean_setting(
    "METENOX_HARVEST_REPROCESS_YIELD ", 0.40
)
"""
Yield at which the metenox reprocess the harvested materials.
This value shouldn't be edited
"""

METENOX_FUEL_BLOCKS_PER_HOUR = clean_setting("METENOX_FUEL_BLOCKS_PER_HOUR", 5)
"""
How many fuel blocks a running Metenox consumes every hours.
This value shouldn't be edited
"""

METENOX_MAGMATIC_GASES_PER_HOUR = clean_setting("METENOX_MAGMATIC_GASES_PER_HOUR", 150)
"""
How many magmatic gases a running Metenox consumes every hours.
This value shouldn't be edited
"""

METENOX_DAYS_BETWEEN_GOO_BAY_PINGS = clean_setting(
    "METENOX_DAYS_BETWEEN_GOO_BAY_PINGS", 2
)
"""
How many days minimum there should be between two ping that the moon material bay is too full
"""



"""Helpers for views."""

from moonmining.models import Moon

from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.admin import User
from allianceauth.services.hooks import get_extension_logger
from app_utils.views import BootstrapStyle

from metenox.models import HoldingCorporation, Metenox

logger = get_extension_logger(__name__)


def generic_details_button_html(
    moon,
    django_detail_view,
    tooltip,
    detail_modal_id,
    fa_icon,
) -> str:
    """
    Return HTML to render a button that will call a modal with an object details.

    Rewriting the function `fontawesome_modal_button_html` from app_utils
    """

    ajax_url = reverse(django_detail_view, args=[moon.pk])

    return format_html(
        '<button type="button" '
        'class="btn btn-{}" '
        'data-bs-toggle="modal" '
        'data-bs-target="#{}" '
        "{}"
        "{}>"
        '<i class="{}"></i>'
        "</button>",
        BootstrapStyle(BootstrapStyle.DEFAULT),
        detail_modal_id,
        mark_safe(f'title="{tooltip}" ') if tooltip else "",
        mark_safe(f'data-ajax_url="{ajax_url}" ') if ajax_url else "",
        fa_icon,
    )


def moon_details_button_html(moon: Moon) -> str:
    """
    Return HTML to render a moon details button.

    Rewriting the function `fontawesome_modal_button_html` from app_utils
    """

    return generic_details_button_html(
        moon, "metenox:moon_details", "Moon details", "modalMoonDetails", "fas fa-moon"
    )


def metenox_details_button_html(metenox: Metenox) -> str:
    """
    Return an HTML button to display the modal with metenox details
    """

    return generic_details_button_html(
        metenox,
        "metenox:metenox_details",
        "Metenox details",
        "modalMetenoxDetails",
        "fas fa-moon",
    )


def corporation_breakdown_button_html(holding: HoldingCorporation) -> str:
    """
    Returns an HTML button to display the modal with corporation tag breakdown
    """

    return generic_details_button_html(
        holding,
        "metenox:breakdown",
        _("Profit breakdown by tags"),
        "modalCorporationBreakdown",
        "fa-solid fa-magnifying-glass",
    )


def corporation_notifications_button_html(holding: HoldingCorporation) -> str:
    """
    Returns an HTML button to display the modal with corporation notifications details
    """

    return generic_details_button_html(
        holding,
        "metenox:notifications",
        "Notification tool",
        "modalCorporationNotifications",
        "fas fa-bell",
    )


def user_has_owner_in_corp(user: User, corporation: HoldingCorporation) -> bool:
    """
    This function will check if the user has an owner in the corporation passed as a parameter.
    Will return true if they have an owner in the corporation. False otherwise
    """
    res = corporation.owners.filter(character_ownership__user=user).exists()
    perm = user.has_perm("metenox.auditor")
    if not (res or perm):
        logger.warning(
            "User id %s tried to access corporation id %s without authorization",
            user.id,
            corporation.corporation_id,
        )

    return res or perm



"""Interactions with the fuzzwork market API"""

from enum import Enum
from typing import Dict, List

import requests

from eveuniverse.models import EveType

from allianceauth.services.hooks import get_extension_logger

from metenox import __version__, repo_url

FUZZWORK_URL = "https://market.fuzzwork.co.uk/aggregates/"
THE_FORGE_REGION_ID = 10000002

logger = get_extension_logger(__name__)


class BuySell(Enum):
    """Parameter to see if you want to fetch buy or sell orders"""

    BUY = "buy"
    SELL = "sell"


class PriceType(Enum):
    """Parameter to know what price aggregate you're looking for"""

    WEIGHTED_AVERAGE = "weightedAverage"
    MAX = "max"
    MIN = "min"
    MEDIAN = "median"
    FIVE_PERCENT_WEIGHTED_AVERAGE = "percentile"


def get_types_prices(
    eve_types: List[EveType], buy_sell: BuySell, price_type: PriceType
) -> Dict[int, float]:
    """
    Retrieves the price of the given eve_types from the Fuzzwork API
    The parameters allow to select the  price type you want
    """

    type_ids = [eve_type.id for eve_type in eve_types]
    return get_type_ids_prices(type_ids, buy_sell, price_type)


def get_type_ids_prices(
    eve_types_ids: List[int],
    buy_sell: BuySell = BuySell.BUY,
    price_type: PriceType = PriceType.FIVE_PERCENT_WEIGHTED_AVERAGE,
) -> Dict[int, float]:
    """
    Retrieves the price of the given eve_type_ids from the Fuzzwork API
    The parameters allow to select the  price type you want
    """

    if not eve_types_ids:
        logger.warning("Received an empty list for fetching prices")
        return {}

    url = f"{FUZZWORK_URL}?region={THE_FORGE_REGION_ID}&types={','.join([str(eve_type_id) for eve_type_id in eve_types_ids])}"

    user_agent = f"aa-metenox v{__version__} {repo_url}"

    logger.info("Trying to fetch data from %s" % url)
    r = requests.get(url, headers={"User-Agent": user_agent}, timeout=5)
    r.raise_for_status()

    output_dic = {}

    for type_id, market_info in r.json().items():
        output_dic[int(type_id)] = float(market_info[buy_sell.value][price_type.value])

    return output_dic



from eveuniverse.models import EveMoon

from allianceauth.eveonline.models import EveCorporationInfo

from metenox.models import HoldingCorporation, Metenox
from metenox.tasks import create_metenox

MOON_ID = 40178441


def create_test_holding(holding_id: int = 1) -> HoldingCorporation:
    """Creates a template holding corporation"""

    corporation = EveCorporationInfo.objects.create(
        corporation_id=holding_id,
        corporation_name="corporation1",
        corporation_ticker="CORP1",
        member_count=1,
    )
    holding = HoldingCorporation(corporation=corporation)
    holding.save()

    return holding


def create_test_metenox() -> Metenox:
    """
    Creates a basic metenox for testing purpose
    """

    corporation = EveCorporationInfo.objects.create(
        corporation_id=1,
        corporation_name="corporation1",
        corporation_ticker="CORP1",
        member_count=1,
    )
    holding = HoldingCorporation(corporation=corporation)
    holding.save()
    eve_moon = EveMoon.objects.get(id=MOON_ID)

    structure_info = {
        "name": "Metenox1",
        "structure_id": 1,
    }

    location_info = {
        "position": {
            "x": eve_moon.position_x,
            "y": eve_moon.position_y,
            "z": eve_moon.position_z,
        },
        "solar_system_id": eve_moon.eve_planet.eve_solar_system.id,
    }

    create_metenox(holding.corporation.corporation_id, structure_info, location_info)

    metenox = Metenox.objects.get(structure_id=1)

    return metenox



# Generated by Django 4.2.15 on 2024-10-07 11:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("metenox", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="MetenoxTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=20)),
            ],
        ),
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("view_moons", "Can see all scanned moons"),
                    (
                        "view_metenoxes",
                        "Can add owners and view related corporations metenoxes",
                    ),
                    ("corporation_manager", "Can modify a corporation ping settings"),
                    (
                        "auditor",
                        "Can access metenox information about all corporations",
                    ),
                ),
            },
        ),
        migrations.AddField(
            model_name="holdingcorporation",
            name="ping_on_remaining_fuel_days",
            field=models.IntegerField(
                default=0,
                help_text="Ping should be sent when the fuel blocks stored in a Metenox allow less than this value",
            ),
        ),
        migrations.AddField(
            model_name="holdingcorporation",
            name="ping_on_remaining_magmatic_days",
            field=models.IntegerField(
                default=0,
                help_text="Ping should be sent when the magmatic gases stored in a Metenox allow less than this value",
            ),
        ),
        migrations.AddField(
            model_name="metenox",
            name="was_fuel_pinged",
            field=models.BooleanField(
                default=False,
                help_text="If a ping has been sent out after noticing a low fuel level",
            ),
        ),
        migrations.AddField(
            model_name="metenox",
            name="was_magmatic_pinged",
            field=models.BooleanField(
                default=False,
                help_text="If a ping has been sent out after noticing a low magmatic level",
            ),
        ),
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "webhook_id",
                    models.BigIntegerField(primary_key=True, serialize=False),
                ),
                ("webhook_token", models.TextField()),
                (
                    "name",
                    models.CharField(
                        help_text="Text to recognize the webhook",
                        max_length=30,
                        unique=True,
                    ),
                ),
                (
                    "default",
                    models.BooleanField(
                        default=False,
                        help_text="If the webhook will automatically be added to every new corporation",
                    ),
                ),
                (
                    "holding_corporations",
                    models.ManyToManyField(
                        help_text="Corporation associated to this webhook",
                        related_name="webhooks",
                        to="metenox.holdingcorporation",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="metenox",
            name="tag",
            field=models.ForeignKey(
                default=None,
                help_text="Tag assigned to a Metenox",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="metenox.metenoxtag",
            ),
        ),
    ]



"""Hooks for charlink application"""

from charlink.app_imports.utils import (  # pylint: disable=import-error
    AppImport,
    LoginImport,
)

from django.contrib import messages
from django.contrib.auth.models import Permission, User
from django.db.models import Exists, OuterRef
from django.shortcuts import get_object_or_404

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.allianceauth import notify_admins, users_with_permission

from metenox import tasks
from metenox.app_settings import METENOX_ADMIN_NOTIFICATIONS_ENABLED
from metenox.models import ESI_SCOPES, HoldingCorporation, Owner, Webhook


# pylint: disable=duplicate-code
def _add_character(request, token):
    character_ownership = get_object_or_404(
        request.user.character_ownerships.select_related("character"),
        character__character_id=token.character_id,
    )
    corporation_id = character_ownership.character.corporation_id
    try:
        corporation = EveCorporationInfo.objects.get(corporation_id=corporation_id)
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            corp_id=corporation_id
        )
        corporation.save()

    holding_corporation, created_corporation = HoldingCorporation.objects.get_or_create(
        corporation=corporation,
    )

    if created_corporation:
        Webhook.add_default_webhooks_to_corporation(holding_corporation)

    owner, created_owner = Owner.objects.get_or_create(
        corporation=holding_corporation, character_ownership=character_ownership
    )
    if not created_owner:
        owner.enable()  # Gives another chance to the toon at being used for updates

    # TODO figure out why I need to type all this to get the right corp id
    tasks.update_holding.delay(owner.corporation.corporation.corporation_id)
    messages.success(request, f"Update of refineries started for {owner}.")
    if METENOX_ADMIN_NOTIFICATIONS_ENABLED:
        notify_admins(
            message=f"{owner} was added as new owner by {request.user}.",
            title=f"Metenox: Owner added: {owner}",
        )


def _check_permissions(user: User) -> bool:
    """Checks if the user can use the Metenox application"""

    return user.has_perm("metenox.view_moons") or user.has_perm(
        "metenox.view_metenoxes"
    )


def _is_character_added(character: EveCharacter) -> bool:
    """Checks if the character is already added in the Metenox application"""

    return Owner.objects.filter(character_ownership__character=character).exists()


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(content_type__app_label="metenox", codename="view_moons")
    ) | users_with_permission(
        Permission.objects.get(
            content_type__app_label="metenox", codename="view_metenoxes"
        )
    )


app_import = AppImport(
    "metenox",
    [
        LoginImport(
            app_label="metenox",
            unique_id="default",
            field_label="Metenox",
            add_character=_add_character,
            scopes=ESI_SCOPES,
            check_permissions=_check_permissions,
            is_character_added=_is_character_added,
            is_character_added_annotation=Exists(
                Owner.objects.filter(character_ownership__character_id=OuterRef("pk"))
            ),
            get_users_with_perms=_users_with_perms,
        )
    ],
)



# Generated by Django 4.2.16 on 2025-02-28 18:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("metenox", "0004_alter_metenox_tags"),
    ]

    operations = [
        migrations.AddField(
            model_name="metenox",
            name="last_updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="metenox",
            name="tags",
            field=models.ManyToManyField(
                blank=True,
                help_text="Tags assigned to a Metenox",
                related_name="metenoxes",
                to="metenox.metenoxtag",
            ),
        ),
        migrations.AlterField(
            model_name="metenoxtag",
            name="name",
            field=models.CharField(max_length=20, unique=True),
        ),
    ]



"""
Modules containing all ESI interactions
"""

from collections import defaultdict
from typing import Dict, List, Set

from bravado.exception import HTTPForbidden

from esi.clients import EsiClientProvider
from esi.models import Token

from allianceauth.services.hooks import get_extension_logger
from app_utils.esi import fetch_esi_status

from metenox.models import HoldingCorporation

from . import __version__

METENOX_TYPE_ID = 81826

logger = get_extension_logger(__name__)

esi = EsiClientProvider(app_info_text=f"aa-metenox v{__version__}")


class ESIError(Exception):
    """Signifies that something went wrong when querying data from the ESI"""


class DownTimeError(Exception):
    """Signifies that it is currently the downtime and no data will be returned"""


def get_metenox_from_esi(
    holding_corporation: HoldingCorporation,
) -> List[Dict]:
    """Returns all metenoxes associated with a given Owner"""

    structures = get_structures_from_esi(holding_corporation)

    return [
        structure for structure in structures if structure["type_id"] == METENOX_TYPE_ID
    ]


def get_structure_info_from_esi(
    holding_corporation: HoldingCorporation, structure_id: int
) -> Dict:
    """Returns the location information of a structure"""

    for owner in holding_corporation.owners.all():

        structure_info = esi.client.Universe.get_universe_structures_structure_id(
            structure_id=structure_id,
            token=owner.fetch_token().valid_access_token(),
        ).result()

        return structure_info


def get_structures_from_esi(
    holding_corporation: HoldingCorporation,
) -> List[Dict]:
    """Returns all structures associated with a given owner"""

    if fetch_esi_status().is_daily_downtime:
        raise DownTimeError

    for owner in holding_corporation.active_owners():
        try:
            return esi.client.Corporation.get_corporations_corporation_id_structures(
                corporation_id=owner.corporation.corporation.corporation_id,
                token=owner.fetch_token().valid_access_token(),
            ).results()
        except Token.DoesNotExist:
            logger.error("No token found for owner %s when fetching assets", owner)
            owner.disable(
                cause="ESI error fetching assets. No token found for this character."
            )
        except HTTPForbidden as e:
            logger.error(
                "HTTPForbidden error when fetching holding corporation %s assets with owner %s. Error: %s",
                holding_corporation,
                owner,
                e,
            )
            owner.disable(
                cause="ESI error fetching assets. The character might not be a director."
            )
        except OSError as e:
            logger.warning(
                "Unexpected OsError when fetching holding corporation %s assets with owner %s. Error: %s",
                holding_corporation,
                owner,
                e,
            )

        continue

    raise ESIError(
        "All active owners returned exceptions when trying to get structure data"
    )


def get_corporation_assets(holding_corporation: HoldingCorporation):
    """Returns all the assets of a corporation"""

    if fetch_esi_status().is_daily_downtime:
        raise DownTimeError

    for owner in holding_corporation.active_owners():
        try:
            return esi.client.Assets.get_corporations_corporation_id_assets(
                corporation_id=holding_corporation.corporation.corporation_id,
                token=owner.fetch_token().valid_access_token(),
            ).results()
        except Token.DoesNotExist:
            logger.error("No token found for owner %s when fetching assets", owner)
            owner.disable(
                cause="ESI error fetching assets. No token found for this character."
            )
        except HTTPForbidden as e:
            logger.error(
                "HTTPForbidden error when fetching holding corporation %s assets with owner %s. Error: %s",
                holding_corporation,
                owner,
                e,
            )
            owner.disable(
                cause="ESI error fetching assets. The character might not be a director."
            )
        except OSError as e:
            logger.warning(
                "Unexpected OsError when fetching holding corporation %s assets with owner %s. Error: %s",
                holding_corporation,
                owner,
                e,
            )

        continue

    raise ESIError(
        "All active owners returned exceptions when trying to get their structure data"
    )


def get_corporation_metenox_assets(
    holding_corporation: HoldingCorporation, metenoxes_set_ids: Set[int]
) -> Dict[int, List[Dict]]:
    """
    Return the assets in the corporation's Metenoxes' MoonMaterialBay and FuelBay.
    Need to receive the set of the corporation's metenoxes ids.
    The data is formatted as a dict with the key being the metenox structure id and a list with the info
    """
    logger.info(
        "Requesting metenox assets of corporation id %d",
        holding_corporation.corporation_id,
    )

    interesting_location_flags = [
        "MoonMaterialBay",
        "StructureFuel",
    ]

    holding_assets = get_corporation_assets(holding_corporation)
    logger.debug("Raw ESI assets: %s", holding_assets)
    assets_dic = defaultdict(list)
    for asset in holding_assets:
        if (
            asset["location_id"] in metenoxes_set_ids
            and asset["location_flag"] in interesting_location_flags
        ):
            assets_dic[asset["location_id"]].append(asset)

    logger.debug("Resulting metenox assets: %s", assets_dic)

    return assets_dic



from unittest.mock import MagicMock

from moonmining.models import Moon as MoonMiningMoon
from moonmining.models import MoonProduct

from django.test import TestCase
from eveuniverse.models import EveType

from metenox.models import (
    EveTypePrice,
    HoldingCorporation,
    MetenoxStoredMoonMaterials,
    MetenoxTag,
)
from metenox.tasks import update_moons_from_moonmining
from metenox.tests.testdata.load_eveuniverse import load_eveuniverse
from metenox.tests.utils import MOON_ID, create_test_metenox


class TestMetenoxes(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_moon_update_when_metenox_created_on_empty_moon(self):
        """
        Check that when a metenox is created on a moon that isn't present in moonmining an update will happen
        if the moon appears in moonmining
        """

        metenox = create_test_metenox()

        self.assertEqual(
            len(metenox.moon.hourly_pull), 0
        )  # checks that the moon is empty
        self.assertEqual(metenox.moon.moonmining_moon, None)

        moon_scan = {
            45490: 0.2809781134,
            45491: 0.2083230466,
            45493: 0.5106988549,
        }

        moonmining_moon, _ = MoonMiningMoon.objects.get_or_create(eve_moon_id=MOON_ID)

        moon_products = [
            MoonProduct(
                moon=moonmining_moon, ore_type_id=ore_type_id, amount=ore_type_amount
            )
            for ore_type_id, ore_type_amount in moon_scan.items()
        ]
        moonmining_moon.update_products(moon_products)

        update_moons_from_moonmining()

        self.assertEqual(
            len(metenox.moon.hourly_pull), 3
        )  # checks that  the moon was updated

    def test_total_stored_volume(self):
        """Checks the volume stored in a metenox"""

        metenox = create_test_metenox()

        atmospheric_gases = EveType.objects.get(id=16634)
        hydrocarbons = EveType.objects.get(id=16633)

        stored_atmospheric_gases = MetenoxStoredMoonMaterials(
            metenox=metenox, product=atmospheric_gases, amount=50
        )
        stored_atmospheric_gases.save()
        stored_hydrocarbons = MetenoxStoredMoonMaterials(
            metenox=metenox, product=hydrocarbons, amount=200
        )
        stored_hydrocarbons.save()

        self.assertEqual(
            metenox.get_stored_moon_materials_volume(), 50 * 0.05 + 200 * 0.05
        )

    def test_total_stored_value(self):
        """Checks the value stored in a metenox"""

        metenox = create_test_metenox()
        hydrocarbon = EveType.objects.get(id=16633)
        atmospheric_gases = EveType.objects.get(id=16634)
        EveTypePrice.objects.create(eve_type=hydrocarbon, price=2000)
        EveTypePrice.objects.create(eve_type=atmospheric_gases, price=1000)

        MetenoxStoredMoonMaterials.objects.create(
            metenox=metenox, product=atmospheric_gases, amount=50
        )
        MetenoxStoredMoonMaterials.objects.create(
            metenox=metenox, product=hydrocarbon, amount=200
        )

        self.assertEqual(
            2000 * 200 + 1000 * 50, metenox.get_stored_moon_materials_value()
        )

    def test_set_fuel_blocks(self):
        """Test the set_fuel_blocks command"""

        metenox = create_test_metenox()
        holding = metenox.corporation
        holding.ping_on_remaining_fuel_days = (
            1  # 120 fuel blocks being a daily consumption
        )
        holding.save()

        self.assertEqual(metenox.fuel_blocks_count, 0)

        metenox.set_fuel_blocs(1400)

        self.assertEqual(metenox.fuel_blocks_count, 1400)
        self.assertFalse(metenox.was_fuel_pinged)  # enough fuel to last longer

        metenox.set_fuel_blocs(20)

        self.assertEqual(metenox.fuel_blocks_count, 20)
        self.assertTrue(metenox.was_fuel_pinged)

        metenox.set_fuel_blocs(1500)

        self.assertEqual(metenox.fuel_blocks_count, 1500)
        self.assertFalse(metenox.was_fuel_pinged)

    def test_set_magmatic(self):
        """Test the set_magmatic command"""

        metenox = create_test_metenox()
        holding = metenox.corporation
        holding.ping_on_remaining_magmatic_days = (
            1  # 2640 magmatic gases being a daily consumption
        )
        holding.save()

        self.assertEqual(metenox.fuel_blocks_count, 0)

        metenox.set_magmatic_gases(4000)

        self.assertEqual(metenox.magmatic_gas_count, 4000)
        self.assertFalse(metenox.was_magmatic_pinged)

        metenox.set_magmatic_gases(1000)

        self.assertEqual(metenox.magmatic_gas_count, 1000)
        self.assertTrue(metenox.was_magmatic_pinged)

        metenox.set_magmatic_gases(5000)

        self.assertEqual(metenox.magmatic_gas_count, 5000)
        self.assertFalse(metenox.was_magmatic_pinged)

    def test_default_tags(self):
        """
        Checks that on metenox creation the default tags are correctly added
        """

        tag1 = MetenoxTag.objects.create(name="tag1", default=True)
        tag2 = MetenoxTag.objects.create(name="tag2", default=True)
        tag3 = MetenoxTag.objects.create(name="tag3")

        metenox = create_test_metenox()

        self.assertIn(tag1, metenox.tags.all())
        self.assertIn(tag2, metenox.tags.all())
        self.assertNotIn(tag3, metenox.tags.all())

    def test_check_moon_material_bay(self):
        HoldingCorporation.ping_metenox_value = MagicMock()
        HoldingCorporation.ping_metenox_volume = MagicMock()

        metenox = create_test_metenox()
        holding = metenox.corporation
        scandium = EveType.objects.get(id=16639)  # volume = 0.05

        scandium_price = EveTypePrice.objects.create(eve_type=scandium, price=100)
        holding.ping_on_value_in_moongoo_bay = 1000
        holding.save()

        MetenoxStoredMoonMaterials.objects.create(
            metenox=metenox,
            product=scandium,
            amount=1,
        )

        # nothing should ping
        metenox.check_moon_material_bay()

        HoldingCorporation.ping_metenox_value.assert_not_called()
        HoldingCorporation.ping_metenox_volume.assert_not_called()

        metenox.last_moongoo_bay_ping = None  # Resets timer
        metenox.save()

        MetenoxStoredMoonMaterials.objects.update_or_create(
            metenox=metenox, product=scandium, defaults={"amount": 100}
        )

        # value should ping
        metenox.check_moon_material_bay()

        HoldingCorporation.ping_metenox_value.assert_called_once()
        HoldingCorporation.ping_metenox_volume.assert_not_called()

        HoldingCorporation.ping_metenox_value = MagicMock()  # Resets call count
        metenox.last_moongoo_bay_ping = None  # Resets timer
        metenox.save()

        holding.ping_on_volume_in_moongoo_bay = 1  # volume should be 2 so ping

        # value and volume should ping
        metenox.check_moon_material_bay()

        HoldingCorporation.ping_metenox_value.assert_called_once()
        HoldingCorporation.ping_metenox_volume.assert_called_once()

        # resets calls
        HoldingCorporation.ping_metenox_value = MagicMock()
        HoldingCorporation.ping_metenox_volume = MagicMock()
        metenox.last_moongoo_bay_ping = None  # Resets timer
        metenox.save()

        scandium_price.price = 0  # value shouldn't ping
        scandium_price.save()

        # volume should ping
        metenox.check_moon_material_bay()

        HoldingCorporation.ping_metenox_value.assert_not_called()
        HoldingCorporation.ping_metenox_volume.assert_called_once()



# Generated by Django 4.2.16 on 2025-03-12 23:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("metenox", "0005_metenox_last_updated_alter_metenox_tags_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="holdingcorporation",
            name="ping_on_value_in_moongoo_bay",
            field=models.IntegerField(
                blank=True,
                help_text="Ping should be sent when the value stored in the moon bay of a Metenox exceed this value",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="holdingcorporation",
            name="ping_on_volume_in_moongoo_bay",
            field=models.IntegerField(
                blank=True,
                help_text="Ping should be sent when the volume stored in the moon goo bay of a Metenox exceed this value.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="metenox",
            name="last_moongoo_bay_ping",
            field=models.DateTimeField(
                help_text="Last time that a ping was sent about the moon goo bay exceeding the threshold",
                null=True,
            ),
        ),
    ]



# Generated by Django 4.2.15 on 2024-10-08 22:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("metenox", "0002_metenoxtag_alter_general_options_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="metenox",
            name="tag",
        ),
        migrations.AddField(
            model_name="metenox",
            name="tags",
            field=models.ManyToManyField(
                help_text="Tags assigned to a Metenox", to="metenox.metenoxtag"
            ),
        ),
        migrations.AddField(
            model_name="metenoxtag",
            name="default",
            field=models.BooleanField(
                default=False,
                help_text="This tag will be applied on every new metenox if true",
            ),
        ),
    ]



"""Template Tags."""

import datetime as dt
from typing import Optional

from moonmining.constants import DATETIME_FORMAT

from django import template
from eveuniverse.models import EveType

from metenox import models

register = template.Library()


@register.filter
def formatisk(value, magnitude: str = None) -> Optional[str]:
    """Return the formatted ISK value or None if input was invalid.

    Args:
    - magnitude: use the given magnitude to format the number, e.g. "b"
    """
    try:
        value = float(value)
    except (ValueError, TypeError):
        return None
    negative = False
    if value < 0:
        value *= -1
        negative = True
    power_map = {"t": 12, "b": 9, "m": 6, "k": 3, "": 0}
    if magnitude not in power_map:
        if value >= 10**12:
            magnitude = "t"
        elif value >= 10**9:
            magnitude = "b"
        elif value >= 10**6:
            magnitude = "m"
        elif value >= 10**3:
            magnitude = "k"
        else:
            magnitude = ""
    result = f"{value / 10 ** power_map[magnitude]:,.1f}{magnitude}"
    if negative:
        result = f"-{result}"
    return result


@register.simple_tag
def goo_price(
    moon_goo_type: EveType, amount: int = 1, monthly: bool = False
) -> Optional[str]:
    """
    Returns the stored price of a moon goo
    If amount is mentioned it will return the price of `amount` units of goo
    If monthly is added it will multiply the price by `24 * 30` to mimic the value of `amount` hourly for a month
    """
    if value := models.EveTypePrice.get_eve_type_price(moon_goo_type):
        if monthly:
            amount *= 24 * 30
        return formatisk(value * amount)
    return None


@register.filter
def datetime(value: dt.datetime) -> Optional[str]:
    """Render as datetime if possible or return None."""
    try:
        return value.strftime(DATETIME_FORMAT)
    except AttributeError:
        return None


@register.filter
def hours_to_days(number: int) -> str:
    """Converts a number of hours to a str indicating number of months/days/hours"""
    n_hours = number % 24
    number = number // 24
    n_days = number % 31
    n_months = number // 31
    if n_months:
        return f"{n_months} months {n_days} days {n_hours} hours"
    return f"{n_days} days {n_hours} hours"
